==> json-any-2023.1.tar.gz: json-any-2023.1/json_any/object_to_json.py <==
# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2022)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import json
from typing import Any, Sequence

from json_any.catalog.module import grph, nmpy
from json_any.catalog.type import (
    NX_GRAPH_CLASSES,
    SPARSE_ARRAY_CLASSES,
    array_t,
    data_frame_t,
    date_t,
    date_time_t,
    enum_t,
    figure_t,
    io_bytes_t,
    named_tuple_t,
    path_t,
    py_array_t,
    series_t,
    time_delta_t,
    time_t,
    time_zone_t,
    uuid_t,
)
from json_any.constant import (
    CUSTOM_PREFIX,
    DATACLASS_PREFIX,
    DESCRIPTION_FOR_JSON,
    JSON_TYPE,
    JSON_TYPE_NUMPY_SCALAR,
    JSON_TYPE_PREFIX_NETWORKX,
    JSON_TYPE_PREFIX_PATHLIB,
    JSON_TYPE_PREFIX_SCIPY,
    JSONING_ERROR_MARKER,
    STANDARD_PREFIX,
    UNHANDLED_PREFIX,
)
from json_any.numpy import AsMostConciseRepresentation
from json_any.type import (
    FullyQualifiedType,
    IsFullyDataclassBased,
    IsNamedTuple,
    TypeOfInstance,
    description_h,
    descriptors_h,
)

_HISTORY_INDENTATION = "    "


def JsonStringOf(
    instance: Any, /, *, descriptors: descriptors_h = None
) -> tuple[str, Sequence[str] | None]:
    """"""
    history = []
    jsonable = _JsonableVersionOf(
        instance, history, "", highest_level_call=True, descriptors=descriptors
    )
    if all(JSONING_ERROR_MARKER not in _elm for _elm in history):
        history = None

    return json.dumps(jsonable), history


def _JsonableVersionOf(
    instance: Any,
    history: list[str],
    history_level: str,
    /,
    *,
    highest_level_call: bool = False,
    descriptors: descriptors_h = None,
) -> description_h:
    """"""
    instance_type = TypeOfInstance(instance)
    fully_qualified = FullyQualifiedType(instance_type)

    if descriptors is None:
        descriptors = {}

    # Test for high-level classes first since they can also be subclasses of standard
    # classes below.
    in_descriptors_full = fully_qualified in descriptors
    # Empty key "" is equivalent to FullyQualifiedType(instance) when instance is the
    # object passed to "JsonStringOf". It allows to avoid to have to call
    # FullyQualifiedType in the first place.
    in_descriptors_empty = highest_level_call and ("" in descriptors)
    if in_descriptors := (in_descriptors_full or in_descriptors_empty):
        DescriptionForJSON = None
    else:
        DescriptionForJSON = getattr(instance_type, DESCRIPTION_FOR_JSON, None)
    if in_descriptors or (DescriptionForJSON is not None):
        json_type = f"{CUSTOM_PREFIX}{fully_qualified}"
        if in_descriptors_full:
            DescriptionForJSON = descriptors[fully_qualified]
        elif in_descriptors_empty:
            DescriptionForJSON = descriptors[""]
        description = DescriptionForJSON(instance)
        history.append(history_level + json_type)
        jsonable = _JsonableVersionOf(
            description,
            history,
            history_level + _HISTORY_INDENTATION,
            descriptors=descriptors,
        )
    elif dtcl.is_dataclass(instance):
        if IsFullyDataclassBased(instance_type):
            json_type = f"{DATACLASS_PREFIX}{fully_qualified}"
            # Do not use dtcl.asdict(instance) since it recurses into dataclass instances which, if they extend a
            # "container" class like list or dict, might lose information.
            as_dict = {
                _fld.name: getattr(instance, _fld.name)
                for _fld in dtcl.fields(instance)
            }
            history.append(f"{history_level}{json_type}[{as_dict.__len__()}]")
            jsonable = _JsonableVersionOf(
                as_dict,
                history,
                history_level + _HISTORY_INDENTATION,
                descriptors=descriptors,
            )
        else:
            json_type = f"{UNHANDLED_PREFIX}{DATACLASS_PREFIX}{fully_qualified}"
            jsonable = None
            history.append(
                f"{history_level}{JSONING_ERROR_MARKER}"
                f"{instance_type.__name__}: Dataclass with inheritance "
                f'without a "{DESCRIPTION_FOR_JSON}" method. Using None.'
            )
    # Must be the first type to be tested (see un-findable modules above)
    elif instance_type in (bytes, bytearray):
        # bytes.hex was initially used in place of bytes.decode
        json_type = instance_type.__name__
        jsonable = instance.decode(encoding="iso-8859-1")
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is complex:
        json_type = complex.__name__
        jsonable = (instance.real, instance.imag)
        history.append(f"{history_level}{json_type}:{jsonable}")
    # Check datetime before date since a datetime is also a date
    elif instance_type is date_time_t:
        json_type = JSON_TYPE[date_time_t]
        history.append(history_level + json_type)
        jsonable = _JsonableVersionOf(
            (instance.date(), instance.timetz()),
            history,
            history_level + _HISTORY_INDENTATION,
        )
    elif instance_type is date_t:
        json_type = JSON_TYPE[date_t]
        jsonable = (instance.year, instance.month, instance.day)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is time_t:
        json_type = JSON_TYPE[time_t]
        history.append(history_level + json_type)
        jsonable = (
            instance.hour,
            instance.minute,
            instance.second,
            instance.microsecond,
            _JsonableVersionOf(
                instance.tzinfo, history, history_level + _HISTORY_INDENTATION
            ),
            instance.fold,
        )
    elif instance_type is time_delta_t:
        json_type = JSON_TYPE[time_delta_t]
        jsonable = (instance.days, instance.seconds, instance.microseconds)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is time_zone_t:
        json_type = JSON_TYPE[time_zone_t]
        history.append(history_level + json_type)
        jsonable = _JsonableVersionOf(
            (instance.utcoffset(None), instance.tzname(None)),
            history,
            history_level + _HISTORY_INDENTATION,
        )
    elif issubclass(instance_type, enum_t):
        json_type = f"{JSON_TYPE[enum_t]}{FullyQualifiedType(instance_type)}"
        history.append(history_level + json_type)
        jsonable = _JsonableVersionOf(
            instance.value,
            history,
            history_level + _HISTORY_INDENTATION,
            descriptors=descriptors,
        )
    elif instance_type is py_array_t:
        json_type, jsonable = JSON_TYPE[py_array_t], (
            instance.tolist(),
            instance.typecode,
        )
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is slice:
        json_type, jsonable = slice.__name__, (
            instance.start,
            instance.stop,
            instance.step,
        )
        history.append(f"{history_level}{json_type}:{jsonable}")
    # Check before looking for tuples since named tuples are subclasses of... tuples
    elif IsNamedTuple(instance):
        json_type = f"{JSON_TYPE[named_tuple_t]}{fully_qualified}"
        as_tuple = tuple(instance)
        history.append(f"{history_level}{json_type}[{as_tuple.__len__()}]")
        jsonable = _JsonableVersionOf(
            as_tuple,
            history,
            history_level + _HISTORY_INDENTATION,
            descriptors=descriptors,
        )
    elif instance_type in (frozenset, list, set, tuple):
        json_type = instance_type.__name__
        history.append(f"{history_level}{json_type}[{instance.__len__()}]")
        jsonable = [
            _JsonableVersionOf(
                _elm,
                history,
                history_level + _HISTORY_INDENTATION,
                descriptors=descriptors,
            )
            for _elm in instance
        ]
    elif instance_type is dict:
        # json does not accept non-str dictionary keys, hence the json.dumps
        json_type = dict.__name__
        history.append(f"{history_level}{json_type}[{instance.__len__()}]")
        jsonable = {
            json.dumps(
                _JsonableVersionOf(
                    _key,
                    history,
                    history_level + _HISTORY_INDENTATION,
                    descriptors=descriptors,
                )
            ): _JsonableVersionOf(
                _vle,
                history,
                history_level + _HISTORY_INDENTATION,
                descriptors=descriptors,
            )
            for _key, _vle in instance.items()
        }
    elif issubclass(instance_type, path_t):
        (
            json_type,
            jsonable,
        ) = f"{JSON_TYPE_PREFIX_PATHLIB}{instance_type.__name__}", str(instance)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is io_bytes_t:
        # Buffer is assumed to be open (i.e. no instance.closed check)
        json_type = JSON_TYPE[io_bytes_t]
        jsonable = instance.getvalue().decode(encoding="iso-8859-1")
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is uuid_t:
        json_type, jsonable = JSON_TYPE[uuid_t], instance.int
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif nmpy.issubclass_(instance_type, (nmpy.bool_, nmpy.number, nmpy.character)):
        json_type = JSON_TYPE_NUMPY_SCALAR
        if nmpy.issubclass_(instance_type, nmpy.bool_):
            value = bool(instance)
        elif nmpy.issubclass_(instance_type, nmpy.integer):
            value = int(instance)
        elif nmpy.issubclass_(instance_type, nmpy.floating):
            value = float(instance)
        elif nmpy.issubclass_(instance_type, complex):
            # Use list (instead of tuple) since tuples seem to be unJSONed as list, so...
            value = [float(nmpy.real(instance)), float(nmpy.imag(instance))]
        else:
            value = str(instance)
        jsonable = (instance.dtype.char, value)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is array_t:
        json_type, jsonable = JSON_TYPE[array_t], AsMostConciseRepresentation(instance)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type in SPARSE_ARRAY_CLASSES:
        (
            json_type,
            jsonable,
        ) = (
            f"{JSON_TYPE_PREFIX_SCIPY}{instance_type.__name__}",
            AsMostConciseRepresentation(instance.toarray()),
        )
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type in (data_frame_t, series_t):
        json_type, jsonable = JSON_TYPE[instance_type], instance.to_json()
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type in NX_GRAPH_CLASSES:
        edges = grph.to_dict_of_dicts(instance)
        # /!\ Node attributes are added to the edges dictionary! This must be taken into account when deJSONing. Note
        # that several attempts to avoid this have been made, including one relying on repr(node), which is based on
        # hash(node). Since the hash function gives different results across Python sessions, this could not work.
        for node, attributes in instance.nodes(data=True):
            edges[node] = (attributes, edges[node])
        json_type = f"{JSON_TYPE_PREFIX_NETWORKX}{instance_type.__name__}"
        history.append(history_level + json_type)
        jsonable = _JsonableVersionOf(
            edges,
            history,
            history_level + _HISTORY_INDENTATION,
            descriptors=descriptors,
        )
    elif instance_type is figure_t:
        fake_file = io_bytes_t()
        instance.canvas.draw()
        instance.savefig(
            fake_file,
            bbox_inches="tight",
            pad_inches=0.0,
            transparent=True,
            dpi=200.0,
            format="png",
        )
        json_type = JSON_TYPE[figure_t]
        jsonable = fake_file.getvalue().decode(encoding="iso-8859-1")
        fake_file.close()
        history.append(history_level + json_type)
    else:
        try:
            _ = json.dumps(instance)
            json_type = f"{STANDARD_PREFIX}{instance_type.__name__}"
            jsonable = instance
            history.append(f"{history_level}{json_type}:{jsonable}")
        except TypeError:
            json_type = f"{UNHANDLED_PREFIX}{instance_type.__name__}"
            jsonable = None
            history.append(
                f"{history_level}{JSONING_ERROR_MARKER}"
                f"{json_type[UNHANDLED_PREFIX.__len__():]}: "
                f"UnJSONable type. Using None."
            )

    return json_type, jsonable

==> json-any-2023.2.tar.gz: json-any-2023.2/json_any/object_to_json.py <==
# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2022)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import json
from typing import Any, Sequence

from json_any.catalog.module import grph, nmpy
from json_any.catalog.type import (
    NX_GRAPH_CLASSES,
    SPARSE_ARRAY_CLASSES,
    array_t,
    data_frame_t,
    date_t,
    date_time_t,
    enum_t,
    figure_t,
    io_bytes_t,
    named_tuple_t,
    path_t,
    py_array_t,
    series_t,
    time_delta_t,
    time_t,
    time_zone_t,
    uuid_t,
)
from json_any.constant import (
    CUSTOM_PREFIX,
    DATACLASS_PREFIX,
    DESCRIPTION_FOR_JSON,
    JSON_TYPE,
    JSON_TYPE_NUMPY_SCALAR,
    JSON_TYPE_PREFIX_NETWORKX,
    JSON_TYPE_PREFIX_PATHLIB,
    JSON_TYPE_PREFIX_SCIPY,
    JSONING_ERROR_MARKER,
    STANDARD_PREFIX,
    UNHANDLED_PREFIX,
)
from json_any.extension.numpy import AsMostConciseRepresentation
from json_any.extension.type import (
    FullyQualifiedType,
    IsFullyDataclassBased,
    IsNamedTuple,
    TypeOfInstance,
    description_h,
    descriptors_h,
)

_HISTORY_INDENTATION = "    "


def JsonStringOf(
    instance: Any, /, *, descriptors: descriptors_h = None
) -> tuple[str, Sequence[str] | None]:
    """"""
    history = []
    jsonable = _JsonableVersionOf(
        instance, history, "", highest_level_call=True, descriptors=descriptors
    )
    if all(JSONING_ERROR_MARKER not in _elm for _elm in history):
        history = None

    return json.dumps(jsonable), history


def _JsonableVersionOf(
    instance: Any,
    history: list[str],
    history_level: str,
    /,
    *,
    highest_level_call: bool = False,
    descriptors: descriptors_h = None,
) -> description_h:
    """"""
    instance_type = TypeOfInstance(instance)
    fully_qualified = FullyQualifiedType(instance_type)

    if descriptors is None:
        descriptors = {}

    # Test for high-level classes first since they can also be subclasses of standard
    # classes below.
    in_descriptors_full = fully_qualified in descriptors
    # Empty key "" is equivalent to FullyQualifiedType(instance) when instance is the
    # object passed to "JsonStringOf". It allows to avoid to have to call
    # FullyQualifiedType in the first place.
    in_descriptors_empty = highest_level_call and ("" in descriptors)
    if in_descriptors := (in_descriptors_full or in_descriptors_empty):
        DescriptionForJSON = None
    else:
        DescriptionForJSON = getattr(instance_type, DESCRIPTION_FOR_JSON, None)
    if in_descriptors or (DescriptionForJSON is not None):
        json_type = f"{CUSTOM_PREFIX}{fully_qualified}"
        if in_descriptors_full:
            DescriptionForJSON = descriptors[fully_qualified]
        elif in_descriptors_empty:
            DescriptionForJSON = descriptors[""]
        description = DescriptionForJSON(instance)
        history.append(history_level + json_type)
        jsonable = _JsonableVersionOf(
            description,
            history,
            history_level + _HISTORY_INDENTATION,
            descriptors=descriptors,
        )
    elif dtcl.is_dataclass(instance):
        if IsFullyDataclassBased(instance_type):
            json_type = f"{DATACLASS_PREFIX}{fully_qualified}"
            # Do not use dtcl.asdict(instance) since it recurses into dataclass instances which, if they extend a
            # "container" class like list or dict, might lose information.
            as_dict = {
                _fld.name: getattr(instance, _fld.name)
                for _fld in dtcl.fields(instance)
            }
            history.append(f"{history_level}{json_type}[{as_dict.__len__()}]")
            jsonable = _JsonableVersionOf(
                as_dict,
                history,
                history_level + _HISTORY_INDENTATION,
                descriptors=descriptors,
            )
        else:
            json_type = f"{UNHANDLED_PREFIX}{DATACLASS_PREFIX}{fully_qualified}"
            jsonable = None
            history.append(
                f"{history_level}{JSONING_ERROR_MARKER}"
                f"{instance_type.__name__}: Dataclass with inheritance "
                f'without a "{DESCRIPTION_FOR_JSON}" method. Using None.'
            )
    # Must be the first type to be tested (see un-findable modules above)
    elif instance_type in (bytes, bytearray):
        # bytes.hex was initially used in place of bytes.decode
        json_type = instance_type.__name__
        jsonable = instance.decode(encoding="iso-8859-1")
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is complex:
        json_type = complex.__name__
        jsonable = (instance.real, instance.imag)
        history.append(f"{history_level}{json_type}:{jsonable}")
    # Check datetime before date since a datetime is also a date
    elif instance_type is date_time_t:
        json_type = JSON_TYPE[date_time_t]
        history.append(history_level + json_type)
        jsonable = _JsonableVersionOf(
            (instance.date(), instance.timetz()),
            history,
            history_level + _HISTORY_INDENTATION,
        )
    elif instance_type is date_t:
        json_type = JSON_TYPE[date_t]
        jsonable = (instance.year, instance.month, instance.day)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is time_t:
        json_type = JSON_TYPE[time_t]
        history.append(history_level + json_type)
        jsonable = (
            instance.hour,
            instance.minute,
            instance.second,
            instance.microsecond,
            _JsonableVersionOf(
                instance.tzinfo, history, history_level + _HISTORY_INDENTATION
            ),
            instance.fold,
        )
    elif instance_type is time_delta_t:
        json_type = JSON_TYPE[time_delta_t]
        jsonable = (instance.days, instance.seconds, instance.microseconds)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is time_zone_t:
        json_type = JSON_TYPE[time_zone_t]
        history.append(history_level + json_type)
        jsonable = _JsonableVersionOf(
            (instance.utcoffset(None), instance.tzname(None)),
            history,
            history_level + _HISTORY_INDENTATION,
        )
    elif issubclass(instance_type, enum_t):
        json_type = f"{JSON_TYPE[enum_t]}{FullyQualifiedType(instance_type)}"
        history.append(history_level + json_type)
        jsonable = _JsonableVersionOf(
            instance.value,
            history,
            history_level + _HISTORY_INDENTATION,
            descriptors=descriptors,
        )
    elif instance_type is py_array_t:
        json_type, jsonable = JSON_TYPE[py_array_t], (
            instance.tolist(),
            instance.typecode,
        )
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is slice:
        json_type, jsonable = slice.__name__, (
            instance.start,
            instance.stop,
            instance.step,
        )
        history.append(f"{history_level}{json_type}:{jsonable}")
    # Check before looking for tuples since named tuples are subclasses of... tuples
    elif IsNamedTuple(instance):
        json_type = f"{JSON_TYPE[named_tuple_t]}{fully_qualified}"
        as_tuple = tuple(instance)
        history.append(f"{history_level}{json_type}[{as_tuple.__len__()}]")
        jsonable = _JsonableVersionOf(
            as_tuple,
            history,
            history_level + _HISTORY_INDENTATION,
            descriptors=descriptors,
        )
    elif instance_type in (frozenset, list, set, tuple):
        json_type = instance_type.__name__
        history.append(f"{history_level}{json_type}[{instance.__len__()}]")
        jsonable = [
            _JsonableVersionOf(
                _elm,
                history,
                history_level + _HISTORY_INDENTATION,
                descriptors=descriptors,
            )
            for _elm in instance
        ]
    elif instance_type is dict:
        # json does not accept non-str dictionary keys, hence the json.dumps
        json_type = dict.__name__
        history.append(f"{history_level}{json_type}[{instance.__len__()}]")
        jsonable = {
            json.dumps(
                _JsonableVersionOf(
                    _key,
                    history,
                    history_level + _HISTORY_INDENTATION,
                    descriptors=descriptors,
                )
            ): _JsonableVersionOf(
                _vle,
                history,
                history_level + _HISTORY_INDENTATION,
                descriptors=descriptors,
            )
            for _key, _vle in instance.items()
        }
    elif issubclass(instance_type, path_t):
        (
            json_type,
            jsonable,
        ) = f"{JSON_TYPE_PREFIX_PATHLIB}{instance_type.__name__}", str(instance)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is io_bytes_t:
        # Buffer is assumed to be open (i.e. no instance.closed check)
        json_type = JSON_TYPE[io_bytes_t]
        jsonable = instance.getvalue().decode(encoding="iso-8859-1")
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is uuid_t:
        json_type, jsonable = JSON_TYPE[uuid_t], instance.int
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif nmpy.issubclass_(instance_type, (nmpy.bool_, nmpy.number, nmpy.character)):
        json_type = JSON_TYPE_NUMPY_SCALAR
        if nmpy.issubclass_(instance_type, nmpy.bool_):
            value = bool(instance)
        elif nmpy.issubclass_(instance_type, nmpy.integer):
            value = int(instance)
        elif nmpy.issubclass_(instance_type, nmpy.floating):
            value = float(instance)
        elif nmpy.issubclass_(instance_type, complex):
            # Use list (instead of tuple) since tuples seem to be unJSONed as list, so...
            value = [float(nmpy.real(instance)), float(nmpy.imag(instance))]
        else:
            value = str(instance)
        jsonable = (instance.dtype.char, value)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type is array_t:
        json_type, jsonable = JSON_TYPE[array_t], AsMostConciseRepresentation(instance)
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type in SPARSE_ARRAY_CLASSES:
        (
            json_type,
            jsonable,
        ) = (
            f"{JSON_TYPE_PREFIX_SCIPY}{instance_type.__name__}",
            AsMostConciseRepresentation(instance.toarray()),
        )
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type in (data_frame_t, series_t):
        json_type, jsonable = JSON_TYPE[instance_type], instance.to_json()
        history.append(f"{history_level}{json_type}:{jsonable}")
    elif instance_type in NX_GRAPH_CLASSES:
        edges = grph.to_dict_of_dicts(instance)
        # /!\ Node attributes are added to the edges dictionary! This must be taken into account when deJSONing. Note
        # that several attempts to avoid this have been made, including one relying on repr(node), which is based on
        # hash(node). Since the hash function gives different results across Python sessions, this could not work.
        for node, attributes in instance.nodes(data=True):
            edges[node] = (attributes, edges[node])
        json_type = f"{JSON_TYPE_PREFIX_NETWORKX}{instance_type.__name__}"
        history.append(history_level + json_type)
        jsonable = _JsonableVersionOf(
            edges,
            history,
            history_level + _HISTORY_INDENTATION,
            descriptors=descriptors,
        )
    elif instance_type is figure_t:
        fake_file = io_bytes_t()
        instance.canvas.draw()
        instance.savefig(
            fake_file,
            bbox_inches="tight",
            pad_inches=0.0,
            transparent=True,
            dpi=200.0,
            format="png",
        )
        json_type = JSON_TYPE[figure_t]
        jsonable = fake_file.getvalue().decode(encoding="iso-8859-1")
        fake_file.close()
        history.append(history_level + json_type)
    else:
        try:
            _ = json.dumps(instance)
            json_type = f"{STANDARD_PREFIX}{instance_type.__name__}"
            jsonable = instance
            history.append(f"{history_level}{json_type}:{jsonable}")
        except TypeError:
            json_type = f"{UNHANDLED_PREFIX}{instance_type.__name__}"
            jsonable = None
            history.append(
                f"{history_level}{JSONING_ERROR_MARKER}"
                f"{json_type[UNHANDLED_PREFIX.__len__():]}: "
                f"UnJSONable type. Using None."
            )

    return json_type, jsonable
